var keyMaster = (factory, map = new Map()) => ({
	has: key => map.has(key),
	get: key => {
		if (!map.has(key) && typeof factory === 'function') {
			map.set(key, factory(key));
		}

		return map.get(key)
	},
	delete: key => map.delete(key),
	set: (key, value) => map.set(key, value),
	getUnderlyingDataStructure: () => map,
});

const getPropertyValuesInOrder = o => Object.getOwnPropertyNames(o).map(key => o[key]);
const assertType = (name, value, expectedType) => {
	const actualType = typeof value;
	if (actualType !== expectedType) {
		throw new Error(`Expected ${ name } to be ${ expectedType } but it was ${ actualType }`)
	}
};

const freshMap = () => keyMaster(() => Object.create(null));

var betterEmitter = function createEmitter(emitter = Object.create(null)) {
	let eventsToListeners = freshMap();

	emitter.on = (event, listener) => {
		assertType(`event`, event, `string`);
		assertType(`listener`, listener, `function`);

		const id = Math.random().toString();
		const listeners = eventsToListeners.get(event);
		listeners[id] = listener;

		return () => {
			delete listeners[id];
		}
	};

	emitter.once = (event, listener) => {
		assertType(`event`, event, `string`);
		assertType(`listener`, listener, `function`);

		const unsubscribe = emitter.on(event, (...args) => {
			listener(...args);
			unsubscribe();
		});

		return unsubscribe
	};

	emitter.emit = (event, ...args) => {
		assertType(`event`, event, `string`);

		const listeners = eventsToListeners.get(event);
		getPropertyValuesInOrder(listeners).forEach(listener => listener(...args));
	};

	emitter.removeAllListeners = () => {
		eventsToListeners = freshMap();
	};

	return emitter
};

const value = (value) => {
    const emitter = betterEmitter({
        get() {
            return value;
        },
        set(newValue) {
            emitter.emit(`dirty`);
            value = newValue;
            emitter.emit(`value`, value);
        },
        locked() {
            return false;
        },
    });
    return addHelpers(emitter);
};
const combine = (dependencies) => {
    let value = resolve(dependencies);
    let locked = false;
    const setDirty = () => {
        locked = true;
        emitter.emit(`dirty`);
    };
    onEventFromAnyDependency(dependencies, `dirty`, setDirty);
    onEventFromAnyDependency(dependencies, `value`, () => {
        if (!locked) {
            setDirty();
        }
        if (allDependenciesAreUnlocked(dependencies)) {
            value = resolve(dependencies);
            locked = false;
            emitter.emit(`value`, value);
        }
    });
    const emitter = betterEmitter({
        get() {
            if (locked) {
                throw new Error(`Unreadable!`);
            }
            return value;
        },
        set() {
            throw new Error(`Combinations may not be set`);
        },
        locked() {
            return locked;
        },
    });
    return emitter;
};
const computed = (dependencies, calculation) => {
    const combined = combine(dependencies);
    const observableValue = value(calculation(combined.get()));
    const onValueChange = dependencyValues => observableValue.set(calculation(dependencyValues));
    combined.on(`value`, onValueChange);
    return Object.assign(addHelpers(observableValue), {
        set() {
            throw new Error(`Transforms may not be set`);
        },
    });
};
const resolve = (dependencies) => Object.entries(dependencies).reduce((previous, [key, dependency]) => ({
    ...previous,
    [key]: dependency.get()
}), {});
const allDependenciesAreUnlocked = (dependencies) => Object.values(dependencies).every(dependency => !dependency.locked());
const onEventFromAnyDependency = (dependencies, event, cb) => {
    Object.values(dependencies).forEach(dependency => {
        dependency.on(event, cb);
    });
};
const addHelpers = (observable) => Object.assign({}, observable, {
    subscribe(cb) {
        const unsubscribe = observable.on(`value`, cb);
        cb(observable.get());
        return unsubscribe;
    },
    map(fn) {
        return computed({ observable }, ({ observable }) => fn(observable));
    },
});

export { computed, value };
